from typing import Literal

from .base import BaseObject


reaction_names = Literal[
    "heart",
    "like",
    "dislike",
    "laught",
    "wonder",
    "fire",
    "sad",
    "angry",
]


class Reactions(BaseObject):
    heart: int
    like: int
    dislike: int
    laught: int
    wonder: int
    fire: int
    sad: int
    angry: int


class Reacted(BaseObject):
    author: reaction_names



from typing import Literal

from pydantic import HttpUrl, UUID4

from .base import BaseObject


player_urls_size_names = Literal[
    "tiny",      # 144
    "lowest",    # 240
    "low",       # 360
    "medium",    # 480
    "high",      # 720
    "full_hd",   # 1080
    "quad_hd",   # 1440
    "ultra_hd",  # 2160
]


class PlayerUrls(BaseObject):
    type: player_urls_size_names
    url: HttpUrl | Literal[""]


class Text(BaseObject):
    type: Literal["text"]
    content: str
    modificator: str


class Smile(BaseObject):
    type: Literal["smile"]
    smallUrl: HttpUrl
    mediumUrl: HttpUrl
    largeUrl: HttpUrl
    name: str
    isAnimated: bool
    id: UUID4


class Link(BaseObject):
    type: Literal["link"]
    content: str
    url: HttpUrl
    explicit: bool


class LinkToVideo(BaseObject):
    type: Literal["video"]
    url: HttpUrl


class FileBase(BaseObject):
    id: UUID4
    url: HttpUrl


class File(FileBase):
    type: Literal["file"]
    complete: bool
    title: str
    size: int


class Audio(FileBase):
    type: Literal["audio_file"]
    complete: bool
    title: str
    size: int

    # TODO may be None if complete false
    duration: int
    album: str
    artist: str
    track: str


class Video(FileBase):
    type: Literal["ok_video"]
    complete: bool
    """Unknown, probably True if video completely processed (could be False if Post.isRecord)"""
    title: str
    """Video title"""
    duration: int
    """Video duration in seconds"""
    width: int
    """Video max width in pixels"""
    height: int
    """Video max height in pixels"""
    playerUrls: list[PlayerUrls]
    """List of video urls for different resolutions"""
    defaultPreview: HttpUrl
    """random frame from video as thumbnail"""
    preview: HttpUrl
    """author thumbnail or defaultPreview"""
    previewId: UUID4 | None
    """author thumbnail image id or None"""
    vid: int
    """ok.ru video id"""
    failoverHost: str
    """Unknown, probably interchangeable host for playerUrls"""


class Image(FileBase):
    type: Literal["image"]
    rendition: Literal["", "teaser_auto_background"]
    width: int | None
    """Could be None if redention=='teaser_auto_background' and """
    height: int | None
    """Could be None if redention=='teaser_auto_background'"""



import asyncio
from time import time
from aiohttp import ClientSession

from boosty.utils.json import dict_to_file, file_to_dict
from boosty.utils.logging import logger


class Auth:
    access_token, refresh_token, expires_at, device_id, headers = None, None, None, None, None
    DEFAULT_USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36"
    """https://techblog.willshouse.com/2012/01/03/most-common-user-agents/"""

    def __init__(
            self,
            auth_file: str = "auth.json",
            user_agent: str = None,
    ):
        self.auth_file = auth_file
        self.user_agent = user_agent or self.DEFAULT_USER_AGENT
        self.load_auth_data()

    def load_auth_data(self):
        try:
            auth_dict = file_to_dict(self.auth_file)
        except FileNotFoundError:
            logger.info(f"No auth file ({self.auth_file}) was found, using blank values (anonymous access mode)")
            auth_dict = {}
        self.access_token = auth_dict.get("access_token")
        self.refresh_token = auth_dict.get("refresh_token")
        self.expires_at = auth_dict.get("expires_at")
        self.device_id = auth_dict.get("device_id")
        self.headers = {"User-Agent": self.user_agent}
        if self.access_token:
            self.headers |= {"Authorization": f"Bearer {self.access_token}"}

    def save_auth_data(self):
        auth_data = {
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "expires_at": self.expires_at,
            "device_id": self.device_id,
        }
        dict_to_file(auth_data, self.auth_file)

    def save_auth_data_dotenv(self, dotenv_file=None):
        import dotenv

        if not dotenv_file:
            dotenv_file = dotenv.find_dotenv()

        print(f".env file found: {dotenv_file}")
        dotenv.load_dotenv(dotenv_file)
        dotenv.set_key(dotenv_file, "ACCESS_TOKEN", self.access_token, "auto")
        dotenv.set_key(dotenv_file, "REFRESH_TOKEN", self.refresh_token, "auto")

    def get_auth_data(self):
        return self.access_token, self.refresh_token, self.expires_at

    async def refresh_auth_data(self, session: ClientSession):
        self.load_auth_data()

        response = await session.post(
            "/oauth/token/",
            data={
                "device_id": self.device_id,
                "device_os": "web",
                "grant_type": "refresh_token",
                "refresh_token": self.refresh_token,
            },
            headers=self.headers)

        response_data = await response.json()
        self.refresh_token = response_data["refresh_token"]
        self.access_token = response_data["access_token"]
        self.expires_at = int(time()) + response_data["expires_in"]

        self.save_auth_data()


async def main():
    async with ClientSession() as session:
        auth = Auth()
        await auth.refresh_auth_data(session)
        access_token, refresh_token, expires = auth.get_auth_data()
    print(f"{access_token , refresh_token , expires = }")


if __name__ == "__main__":
    asyncio.run(main())



from datetime import datetime

from .base import BaseObject
from .users import DonatorUser


class Donator(BaseObject):
    id: int
    bloggerId: int
    targetId: int
    createdAt: datetime
    user: DonatorUser
    amount: int


class DonatorsResponseExtra(BaseObject):
    isLast: bool
    offset: int | None
    """Unknown, example: 10"""


class DonatorsResponse(BaseObject):
    data: list[Donator]
    extra: DonatorsResponseExtra



try:
    from loguru import logger
except ImportError:
    import logging
    logging.basicConfig(
        level=logging.DEBUG,
        # format=(
        #     '%(filename)s:\t'
        #     '%(levelname)s:\t'
        #     '%(funcName)s():\t'
        #     '%(lineno)d:\t'
        #     '%(message)s')
    )
    logger = logging.LoggerAdapter(logging.getLogger("boosty"))



from pathlib import Path

try:
    from orjson import json
    module = "orjson"
except ImportError:
    import json
    module = "json"


def dict_to_file(data: dict, filename: Path | str):
    to_dump = json.dumps(data)
    with open(filename, "wb" if module == "orjson" else "w") as f:
        f.write(to_dump)


def file_to_dict(filename: Path | str) -> dict:
    with open(filename, "rb" if module == "orjson" else "r") as f:
        return json.loads(f.read())



from typing import Literal

from pydantic import HttpUrl, EmailStr, conint

from .base import BaseObject


class BaseUser(BaseObject):
    id: int
    name: str
    nick: str
    hasAvatar: bool
    avatarUrl: HttpUrl | Literal[""]


class BlogUser(BaseUser):
    blogUrl: str


class DonatorUser(BaseUser):
    nickColor: conint(ge=0, le=15)
    """color id from 0 to 15"""
    displayName: str
    vkplayProfileLink: HttpUrl | None
    email: EmailStr


class Commentator(BaseUser):
    nickColor: conint(ge=0, le=15)
    """color id from 0 to 15"""
    displayName: str
    vkplayProfileLink: HttpUrl | Literal[""]



from typing import Literal, Annotated
from datetime import datetime

from pydantic import UUID4, Field

from .base import BaseObject
from .reply import RepliesResponse
from .users import Commentator
from .reactions import Reactions, Reacted
from .media_types import Text, Smile, Link, Image

CommentData = Annotated[
    Text | Smile | Image | Link,
    Field(discriminator="type")]


class Smile(BaseObject):
    type: Literal["smile"]


class Comment(BaseObject):
    id: UUID4
    intId: int
    createdAt: datetime
    updatedAt: datetime | None
    isBlocked: bool
    isUpdated: bool
    isDeleted: bool
    author: Commentator
    reactions: Reactions
    reacted: Reacted | None
    replyCount: int
    replies: RepliesResponse
    post: dict[Literal["id"], UUID4]
    data: list[CommentData]


class CommentsResponseExtra(BaseObject):
    isLast: bool
    isFirst: bool
    # TODO offset: int | None


class CommentsResponse(BaseObject):
    data: list[Comment]
    extra: CommentsResponseExtra



# stolen from vkbottle with love <3

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, Type, TypeVar

from aiohttp import ClientSession

from boosty.utils.json import json

if TYPE_CHECKING:
    from aiohttp import ClientResponse


TSingleAiohttpClient = TypeVar("TSingleAiohttpClient", bound="SingleAiohttpClient")


class ABCHTTPClient(ABC):
    """Abstract class for http-clients
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/http/http-client.md
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    async def request_raw(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> Any:
        pass

    @abstractmethod
    async def request_text(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> str:
        pass

    @abstractmethod
    async def request_json(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> dict:
        pass

    @abstractmethod
    async def request_content(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> bytes:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass

    async def __aenter__(self) -> "ABCHTTPClient":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()


class AiohttpClient(ABCHTTPClient):
    def __init__(
        self,
        session: Optional[ClientSession] = None,
        json_processing_module: Optional[Any] = None,
        optimize: bool = False,
        **session_params,
    ):
        self.json_processing_module = (
            json_processing_module or session_params.pop("json_serialize", None) or json
        )

        if optimize:
            session_params["skip_auto_headers"] = {"User-Agent"}
            session_params["raise_for_status"] = True

        self.session = session

        self._session_params = session_params

    async def request_raw(
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs,
    ) -> "ClientResponse":
        if not self.session:
            self.session = ClientSession(
                json_serialize=self.json_processing_module.dumps,
                **self._session_params,
            )
        async with self.session.request(url=url, method=method, data=data, **kwargs) as response:
            await response.read()
            return response

    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs,
    ) -> dict:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.json(
            encoding="utf-8", loads=self.json_processing_module.loads, content_type=None
        )

    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs,
    ) -> str:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.text(encoding="utf-8")

    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs,
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return response._body

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

    def __del__(self):
        if self.session and not self.session.closed:
            if self.session._connector is not None and self.session._connector_owner:
                self.session._connector.close()
            self.session._connector = None


class SingleAiohttpClient(AiohttpClient):
    __instance__ = None

    def __new__(
        cls: Type[TSingleAiohttpClient], *args: Any, **kwargs: Any
    ) -> TSingleAiohttpClient:
        if cls.__instance__ is None:
            cls.__instance__ = super().__new__(cls, *args, **kwargs)
        return cls.__instance__

    def __aexit__(self, exc_type, exc_val, exc_tb):
        pass  # no need to close session in this case



from pydantic import conint, UUID4, BaseModel

from boosty.api.auth import Auth
from boosty.types import PostsResponse, Post, CommentsResponse
from boosty.utils.client import AiohttpClient, ABCHTTPClient


class Error(BaseModel):
    error: str
    error_description: str


class BoostyError(ValueError):
    pass


class API:
    API_URL = "https://api.boosty.to"

    def __init__(
            self,
            session: ABCHTTPClient = None,
            auth: Auth = None,
    ):
        self.session = session or AiohttpClient()
        self.auth = auth or Auth()

    async def request(self, method: str, params: dict, data: dict = None) -> dict:
        if not params:
            params = {}
        if not data:
            data = {}
        params = {key: value for key, value in params.items() if value is not None}
        data = {key: value for key, value in data.items() if value is not None}
        response = await self.session.request_raw(
            f"{self.API_URL}{method}",
            method="GET",
            params=params,
            data=data,
            headers=self.auth.headers,
        )

        if response.status // 100 == 5:
            raise BoostyError(Error(error_description=str(response.status), error="Server error"))

        response_json = await response.json(
            encoding="utf-8", content_type=None
        )

        if response.status // 100 == 4:
            raise BoostyError(Error(**response_json))

        return response_json

    async def get_posts(
            self,
            name: str,

            limit: conint(ge=1, le=100) = None,  # limit is based on comments count
            offset: str = None,  # "1654884900:923396"
            comments_limit: conint(ge=0) = None,
            reply_limit: int = None,  # idk ~1
    ) -> PostsResponse:
        resp_json = await self.request(
            f"/v1/blog/{name}/post/", params={
                "limit": limit,
                "offset": offset,
                "comments_limit": comments_limit,
                "reply_limit": reply_limit,
            })
        return PostsResponse(**resp_json)

    async def get_post(
            self,
            name: str,
            post_id: UUID4 | str,

            comments_limit: conint(ge=0) = None,
            reply_limit: int = None,  # idk ~1
    ) -> Post:
        resp_json = await self.request(
            f"/v1/blog/{name}/post/{post_id}", params={
                "comments_limit": comments_limit,
                "reply_limit": reply_limit,
            })
        return Post(**resp_json)

    async def get_post_comments(
            self,
            name: str,
            post_id: UUID4 | str,

            offset: str = None,  # "1654884900:923396"
            limit: conint(ge=0) = None,  # ~20
            reply_limit: int = None,  # idk ~1
            order: str = None,  # ~"top"
    ) -> CommentsResponse:
        resp_json = await self.request(
            f"/v1/blog/{name}/post/{post_id}/comment/", params={
                "offset": offset,
                "limit": limit,
                "reply_limit": reply_limit,
                "order": order,
            })
        return CommentsResponse(**resp_json)



import html
from typing import Literal

from pydantic import HttpUrl

from boosty.api import API
from boosty.types import BaseObject
from boosty.types.media_types import Video
from boosty.utils.json import json

size_dict = {"ultra": 6, "quad": 5, "full": 4, "hd": 3, "sd": 2, "low": 1, "lowest": 0, "mobile": -1}
size_names = Literal[
    "mobile",  # 144
    "lowest",  # 240
    "low",     # 360
    "sd",      # 480
    "hd",      # 720
    "full",    # 1080
    "quad",    # 1440
    "ultra",   # 2160
]


class VideoSize(BaseObject):
    name: size_names
    url: HttpUrl
    seekSchema: Literal[3]
    disallowed: Literal[False]


async def get_video_sizes(
        api: API,
        name: str,
        content: Video
) -> list[VideoSize]:
    """
    :param api: API instance
    :param name: str for native referer header
    :param content: Video to get links from
    :return: list of VideoSize, sorted by quality descending
    """
    player_html = await api.session.request_text(
        content.url, headers={"User-Agent": api.auth.user_agent, "referer": f"https://boosty.to/{name}"})
    ind = player_html.find("data-options=")
    if ind == -1:
        raise ValueError("No data-options found in player script")
    do = ind + 14
    video_data_raw = html.unescape(player_html[do:player_html.find("\"", do)])
    video_data = json.loads(video_data_raw)

    sizes_list = json.loads(video_data["flashvars"]["metadata"])["videos"]
    sizes_list = [VideoSize(**size) for size in sizes_list]
    sizes_list.sort(key=lambda x: -size_dict[x.name])
    return sizes_list



from datetime import datetime
from typing import Literal

from pydantic import UUID4

from .base import BaseObject
from .users import Commentator
from .reactions import Reactions, Reacted


class Reply(BaseObject):
    id: UUID4
    intId: int
    createdAt: datetime
    updatedAt: datetime | None
    isBlocked: bool
    isUpdated: bool
    isDeleted: bool
    author: Commentator
    reactions: Reactions
    reacted: Reacted | None
    replyCount: int
    post: dict[Literal["id"], UUID4]
    data: list[dict]

    replyToUser: Commentator
    replyId: str
    parentId: str


class ReplyResponseExtra(BaseObject):
    isLast: bool
    isFirst: bool
    # TODO offset: int | None


class RepliesResponse(BaseObject):
    data: list[Reply]
    extra: ReplyResponseExtra



import os

from pydantic import BaseModel, Extra


schema_strict = bool(os.getenv("DEBUG"))


class BaseObject(BaseModel):
    class Config:
        extra = Extra.forbid if schema_strict else Extra.ignore



from datetime import datetime

from pydantic import UUID4

from .base import BaseObject
from .comment import CommentsResponse
from .content import Content
from .donator import DonatorsResponse
from .reactions import Reactions
from .teaser import Teaser
from .users import BlogUser


class Currency(BaseObject):
    USD: int | float
    RUB: int | float


class Tag(BaseObject):
    id: int
    title: str
    """Tag name"""


class SubscriptionLevel(BaseObject):
    id: int
    createdAt: datetime
    """Creation timestamp"""
    price: int
    changePrice: int | None
    data: list[Teaser]
    """Could be Conent type"""
    deleted: bool
    isArchived: bool
    name: str
    ownerId: int
    """post.user.id"""

    currencyPrices: Currency


class React(BaseObject):
    actor: str


class Count(BaseObject):
    likes: int
    comments: int
    reactions: Reactions


class Post(BaseObject):
    id: UUID4
    createdAt: datetime
    """Creation timestamp"""
    updatedAt: datetime
    """Last update timestamp"""
    publishTime: datetime
    """Publication timestamp"""
    data: list[Content]
    """TODO List of content, attached to post (text included)"""
    count: Count
    """Count of likes, comments, reactions"""
    comments: CommentsResponse
    hasAccess: bool
    """Is post is available for users"""
    isCommentsDenied: bool
    isLiked: bool
    isPublished: bool
    price: int
    """Price to open post"""
    signedQuery: str
    """Query for media fetching"""
    subscriptionLevel: SubscriptionLevel | None
    """Subscription level for non-free posts"""
    tags: list[Tag]
    teaser: list[Teaser]
    title: str
    user: BlogUser

    reacted: React | None
    """Unknown"""
    isWaitingVideo: bool
    """Unknown, probably an indicator, which shows if video is incomplete"""
    currencyPrices: Currency
    """Unknown"""
    isRecord: bool
    """Is post a stream record"""
    donators: DonatorsResponse
    """TODO Unknown"""
    donations: int
    """Amount of donations"""
    int_id: int
    """Unknown, probably post uuid to int"""


class PostsResponseExtra(BaseObject):
    isLast: bool
    """Is last page"""
    offset: str
    """Value for pagination. Example: :code:`1654884900:923396`"""


class PostsResponse(BaseObject):
    data: list[Post]
    extra: PostsResponseExtra



from typing import Annotated

from pydantic import Field

from .media_types import Link, Text, LinkToVideo, File, Audio, Video, Image


Content = Annotated[
    Link | Text | LinkToVideo | File | Audio | Video | Image,
    Field(discriminator="type")]
